@@ -1,38 +1,60 @@
+import asyncio
 import atexit
+import base64
 import collections
 import datetime
-import base64
 import signal
-from textwrap import dedent
-
-from async_generator import asynccontextmanager
-from contextlib import contextmanager
-
-from time import monotonic
-from queue import Empty
-import asyncio
 import typing as t
+from contextlib import asynccontextmanager, contextmanager
+from queue import Empty
+from textwrap import dedent
+from time import monotonic
 
-from traitlets.config.configurable import LoggingConfigurable
-from traitlets import List, Unicode, Bool, Enum, Any, Type, Dict, Integer, default
-
-from nbformat import NotebookNode
-from nbformat.v4 import output_from_msg
 from jupyter_client import KernelManager
 from jupyter_client.client import KernelClient
+from nbformat import NotebookNode
+from nbformat.v4 import output_from_msg
+from traitlets import (
+    Any,
+    Bool,
+    Callable,
+    Dict,
+    Enum,
+    Integer,
+    List,
+    Type,
+    Unicode,
+    default,
+)
+from traitlets.config.configurable import LoggingConfigurable
 
 from .exceptions import (
     CellControlSignal,
+    CellExecutionComplete,
+    CellExecutionError,
     CellTimeoutError,
     DeadKernelError,
-    CellExecutionComplete,
-    CellExecutionError
 )
-from .util import run_sync, ensure_async
 from .output_widget import OutputWidget
+from .util import ensure_async, run_hook, run_sync
 
 
-def timestamp() -> str:
+def timestamp(msg: t.Optional[Dict] = None) -> str:
+    if msg and 'header' in msg:  # The test mocks don't provide a header, so tolerate that
+        msg_header = msg['header']
+        if 'date' in msg_header and isinstance(msg_header['date'], datetime.datetime):
+            try:
+                # reformat datetime into expected format
+                formatted_time = datetime.datetime.strftime(
+                    msg_header['date'], '%Y-%m-%dT%H:%M:%S.%fZ'
+                )
+                if (
+                    formatted_time
+                ):  # docs indicate strftime may return empty string, so let's catch that too
+                    return formatted_time
+            except Exception:
+                pass  # fallback to a local time
+
     return datetime.datetime.utcnow().isoformat() + 'Z'
 
 
@@ -55,7 +77,7 @@ class NotebookClient(LoggingConfigurable):
         ),
     ).tag(config=True)
 
-    timeout_func: t.Any = Any(
+    timeout_func: t.Callable[..., t.Optional[int]] = Any(
         default_value=None,
         allow_none=True,
         help=dedent(
@@ -100,8 +122,9 @@ class NotebookClient(LoggingConfigurable):
         help=dedent(
             """
             If ``False`` (default), when a cell raises an error the
-            execution is stopped and a `CellExecutionError`
-            is raised.
+            execution is stopped and a ``CellExecutionError``
+            is raised, except if the error name is in
+            ``allow_error_names``.
             If ``True``, execution errors are ignored and the execution
             is continued until the end of the notebook. Output from
             exceptions is included in the cell output in both cases.
@@ -109,19 +132,40 @@ class NotebookClient(LoggingConfigurable):
         ),
     ).tag(config=True)
 
+    allow_error_names: t.List[str] = List(
+        Unicode(),
+        help=dedent(
+            """
+            List of error names which won't stop the execution. Use this if the
+            ``allow_errors`` option it too general and you want to allow only
+            specific kinds of errors.
+            """
+        ),
+    ).tag(config=True)
+
     force_raise_errors: bool = Bool(
         False,
         help=dedent(
             """
             If False (default), errors from executing the notebook can be
             allowed with a ``raises-exception`` tag on a single cell, or the
-            ``allow_errors`` configurable option for all cells. An allowed error
-            will be recorded in notebook output, and execution will continue.
-            If an error occurs when it is not explicitly allowed, a
-            `CellExecutionError` will be raised.
-            If True, `CellExecutionError` will be raised for any error that occurs
-            while executing the notebook. This overrides both the
-            ``allow_errors`` option and the ``raises-exception`` cell tag.
+            ``allow_errors`` or ``allow_error_names`` configurable options for
+            all cells. An allowed error will be recorded in notebook output, and
+            execution will continue. If an error occurs when it is not
+            explicitly allowed, a ``CellExecutionError`` will be raised.
+            If True, ``CellExecutionError`` will be raised for any error that occurs
+            while executing the notebook. This overrides the ``allow_errors``
+            and ``allow_error_names`` options and the ``raises-exception`` cell
+            tag.
+            """
+        ),
+    ).tag(config=True)
+
+    skip_cells_with_tag: str = Unicode(
+        'skip-execution',
+        help=dedent(
+            """
+            Name of the cell tag to use to denote a cell that should be skipped.
             """
         ),
     ).tag(config=True)
@@ -227,8 +271,89 @@ class NotebookClient(LoggingConfigurable):
 
     kernel_manager_class: KernelManager = Type(config=True, help='The kernel manager class to use.')
 
-    @default('kernel_manager_class')
-    def _kernel_manager_class_default(self) -> KernelManager:
+    on_notebook_start: t.Optional[t.Callable] = Callable(
+        default_value=None,
+        allow_none=True,
+        help=dedent(
+            """
+            A callable which executes after the kernel manager and kernel client are setup, and
+            cells are about to execute.
+            Called with kwargs ``notebook``.
+            """
+        ),
+    ).tag(config=True)
+
+    on_notebook_complete: t.Optional[t.Callable] = Callable(
+        default_value=None,
+        allow_none=True,
+        help=dedent(
+            """
+            A callable which executes after the kernel is cleaned up.
+            Called with kwargs ``notebook``.
+            """
+        ),
+    ).tag(config=True)
+
+    on_notebook_error: t.Optional[t.Callable] = Callable(
+        default_value=None,
+        allow_none=True,
+        help=dedent(
+            """
+            A callable which executes when the notebook encounters an error.
+            Called with kwargs ``notebook``.
+            """
+        ),
+    ).tag(config=True)
+
+    on_cell_start: t.Optional[t.Callable] = Callable(
+        default_value=None,
+        allow_none=True,
+        help=dedent(
+            """
+            A callable which executes before a cell is executed and before non-executing cells
+            are skipped.
+            Called with kwargs ``cell`` and ``cell_index``.
+            """
+        ),
+    ).tag(config=True)
+
+    on_cell_execute: t.Optional[t.Callable] = Callable(
+        default_value=None,
+        allow_none=True,
+        help=dedent(
+            """
+            A callable which executes just before a code cell is executed.
+            Called with kwargs ``cell`` and ``cell_index``.
+            """
+        ),
+    ).tag(config=True)
+
+    on_cell_complete: t.Optional[t.Callable] = Callable(
+        default_value=None,
+        allow_none=True,
+        help=dedent(
+            """
+            A callable which executes after a cell execution is complete. It is
+            called even when a cell results in a failure.
+            Called with kwargs ``cell`` and ``cell_index``.
+            """
+        ),
+    ).tag(config=True)
+
+    on_cell_error: t.Optional[t.Callable] = Callable(
+        default_value=None,
+        allow_none=True,
+        help=dedent(
+            """
+            A callable which executes when a cell execution results in an error.
+            This is executed even if errors are suppressed with ``cell_allows_errors``.
+            Called with kwargs ``cell` and ``cell_index``.
+            """
+        ),
+    ).tag(config=True)
+
+    @default('kernel_manager_class')  # type:ignore[misc]
+    def _kernel_manager_class_default(self) -> t.Type[KernelManager]:
         """Use a dynamic default to avoid importing jupyter_client at startup"""
         from jupyter_client import AsyncKernelManager
 
@@ -277,11 +402,7 @@ class NotebookClient(LoggingConfigurable):
         )
     )
 
-    def __init__(
-            self,
-            nb: NotebookNode,
-            km: t.Optional[KernelManager] = None,
-            **kw) -> None:
+    def __init__(self, nb: NotebookNode, km: t.Optional[KernelManager] = None, **kw: t.Any) -> None:
         """Initializes the execution manager.
 
         Parameters
@@ -299,9 +420,7 @@ class NotebookClient(LoggingConfigurable):
         self.kc: t.Optional[KernelClient] = None
         self.reset_execution_trackers()
         self.widget_registry: t.Dict[str, t.Dict] = {
-            '@jupyter-widgets/output': {
-                'OutputModel': OutputWidget
-            }
+            '@jupyter-widgets/output': {'OutputModel': OutputWidget}
         }
         # comm_open_handlers should return an object with a .handle_msg(msg) method or None
         self.comm_open_handlers: t.Dict[str, t.Any] = {
@@ -309,17 +428,16 @@ class NotebookClient(LoggingConfigurable):
         }
 
     def reset_execution_trackers(self) -> None:
-        """Resets any per-execution trackers.
-        """
+        """Resets any per-execution trackers."""
         self.task_poll_for_reply: t.Optional[asyncio.Future] = None
         self.code_cells_executed = 0
         self._display_id_map = {}
         self.widget_state: t.Dict[str, t.Dict] = {}
-        self.widget_buffers: t.Dict[str, t.List[t.Dict[str, str]]] = {}
+        self.widget_buffers: t.Dict[str, t.Dict[t.Tuple[str, ...], t.Dict[str, str]]] = {}
         # maps to list of hooks, where the last is used, this is used
         # to support nested use of output widgets.
         self.output_hook_stack: t.Any = collections.defaultdict(list)
-        # our front-end mimicing Output widgets
+        # our front-end mimicking Output widgets
         self.comm_objects: t.Dict[str, t.Any] = {}
 
     def create_kernel_manager(self) -> KernelManager:
@@ -339,6 +457,7 @@ class NotebookClient(LoggingConfigurable):
             self.km = self.kernel_manager_class(config=self.config)
         else:
             self.km = self.kernel_manager_class(kernel_name=self.kernel_name, config=self.config)
+        assert self.km is not None
 
         # If the current kernel manager is still using the default (synchronous) KernelClient class,
         # switch to the async version since that's what NBClient prefers.
@@ -361,14 +480,14 @@ class NotebookClient(LoggingConfigurable):
         finally:
             # Remove any state left over even if we failed to stop the kernel
             await ensure_async(self.km.cleanup_resources())
-            if getattr(self, "kc") and self.kc is not None:
-                await ensure_async(self.kc.stop_channels())
+            if getattr(self, "kc", None) and self.kc is not None:
+                await ensure_async(self.kc.stop_channels())  # type:ignore
                 self.kc = None
                 self.km = None
 
     _cleanup_kernel = run_sync(_async_cleanup_kernel)
 
-    async def async_start_new_kernel(self, **kwargs) -> None:
+    async def async_start_new_kernel(self, **kwargs: t.Any) -> None:
         """Creates a new kernel.
 
         Parameters
@@ -384,12 +503,15 @@ class NotebookClient(LoggingConfigurable):
             kwargs["cwd"] = resource_path
 
         has_history_manager_arg = any(
-            arg.startswith('--HistoryManager.hist_file') for arg in self.extra_arguments)
-        if (hasattr(self.km, 'ipykernel')
-                and self.km.ipykernel
-                and self.ipython_hist_file
-                and not has_history_manager_arg):
-            self.extra_arguments += ['--HistoryManager.hist_file={}'.format(self.ipython_hist_file)]
+            arg.startswith('--HistoryManager.hist_file') for arg in self.extra_arguments
+        )
+        if (
+            hasattr(self.km, 'ipykernel')
+            and self.km.ipykernel
+            and self.ipython_hist_file
+            and not has_history_manager_arg
+        ):
+            self.extra_arguments += [f'--HistoryManager.hist_file={self.ipython_hist_file}']
 
         await ensure_async(self.km.start_kernel(extra_arguments=self.extra_arguments, **kwargs))
 
@@ -405,19 +527,20 @@ class NotebookClient(LoggingConfigurable):
         """
         assert self.km is not None
         self.kc = self.km.client()
-        await ensure_async(self.kc.start_channels())
+        await ensure_async(self.kc.start_channels())  # type:ignore[func-returns-value]
         try:
             await ensure_async(self.kc.wait_for_ready(timeout=self.startup_timeout))
         except RuntimeError:
             await self._async_cleanup_kernel()
             raise
         self.kc.allow_stdin = False
+        await run_hook(self.on_notebook_start, notebook=self.nb)
         return self.kc
 
     start_new_kernel_client = run_sync(async_start_new_kernel_client)
 
     @contextmanager
-    def setup_kernel(self, **kwargs) -> t.Generator:
+    def setup_kernel(self, **kwargs: t.Any) -> t.Generator:
         """
         Context manager for setting up the kernel to execute a notebook.
 
@@ -444,7 +567,7 @@ class NotebookClient(LoggingConfigurable):
                 self._cleanup_kernel()
 
     @asynccontextmanager
-    async def async_setup_kernel(self, **kwargs) -> t.AsyncGenerator:
+    async def async_setup_kernel(self, **kwargs: t.Any) -> t.AsyncGenerator:
         """
         Context manager for setting up the kernel to execute a notebook.
 
@@ -473,7 +596,7 @@ class NotebookClient(LoggingConfigurable):
         try:
             loop.add_signal_handler(signal.SIGINT, on_signal)
             loop.add_signal_handler(signal.SIGTERM, on_signal)
-        except (NotImplementedError, RuntimeError):
+        except RuntimeError:
             # NotImplementedError: Windows does not support signals.
             # RuntimeError: Raised when add_signal_handler is called outside the main thread
             pass
@@ -483,21 +606,21 @@ class NotebookClient(LoggingConfigurable):
             await self.async_start_new_kernel_client()
         try:
             yield
+        except RuntimeError as e:
+            await run_hook(self.on_notebook_error, notebook=self.nb)
+            raise e
         finally:
             if cleanup_kc:
                 await self._async_cleanup_kernel()
-
+            await run_hook(self.on_notebook_complete, notebook=self.nb)
             atexit.unregister(self._cleanup_kernel)
             try:
                 loop.remove_signal_handler(signal.SIGINT)
                 loop.remove_signal_handler(signal.SIGTERM)
-            except (NotImplementedError, RuntimeError):
+            except RuntimeError:
                 pass
 
-    async def async_execute(
-            self,
-            reset_kc: bool = False,
-            **kwargs) -> NotebookNode:
+    async def async_execute(self, reset_kc: bool = False, **kwargs: t.Any) -> NotebookNode:
         """
         Executes each code cell.
 
@@ -564,12 +687,9 @@ class NotebookClient(LoggingConfigurable):
             ]['state'].items():
                 buffers = self.widget_buffers.get(key)
                 if buffers:
-                    widget['buffers'] = buffers
+                    widget['buffers'] = list(buffers.values())
 
-    def _update_display_id(
-            self,
-            display_id: str,
-            msg: t.Dict) -> None:
+    def _update_display_id(self, display_id: str, msg: t.Dict) -> None:
         """Update outputs with a given display_id"""
         if display_id not in self._display_id_map:
             self.log.debug("display id %r not in %s", display_id, self._display_id_map)
@@ -581,7 +701,7 @@ class NotebookClient(LoggingConfigurable):
         try:
             out = output_from_msg(msg)
         except ValueError:
-            self.log.error("unhandled iopub msg: " + msg['msg_type'])
+            self.log.error(f"unhandled iopub msg: {msg['msg_type']}")
             return
 
         for cell_idx, output_indices in self._display_id_map[display_id].items():
@@ -592,12 +712,13 @@ class NotebookClient(LoggingConfigurable):
                 outputs[output_idx]['metadata'] = out['metadata']
 
     async def _async_poll_for_reply(
-            self,
-            msg_id: str,
-            cell: NotebookNode,
-            timeout: t.Optional[int],
-            task_poll_output_msg: asyncio.Future,
-            task_poll_kernel_alive: asyncio.Future) -> t.Dict:
+        self,
+        msg_id: str,
+        cell: NotebookNode,
+        timeout: t.Optional[int],
+        task_poll_output_msg: asyncio.Future,
+        task_poll_kernel_alive: asyncio.Future,
+    ) -> t.Dict:
 
         assert self.kc is not None
         new_timeout: t.Optional[float] = None
@@ -606,10 +727,10 @@ class NotebookClient(LoggingConfigurable):
             new_timeout = float(timeout)
         while True:
             try:
-                msg = await ensure_async(self.kc.shell_channel.get_msg(timeout=new_timeout))
+                msg: t.Dict = await ensure_async(self.kc.shell_channel.get_msg(timeout=new_timeout))
                 if msg['parent_header'].get('msg_id') == msg_id:
                     if self.record_timing:
-                        cell['metadata']['execution']['shell.execute_reply'] = timestamp()
+                        cell['metadata']['execution']['shell.execute_reply'] = timestamp(msg)
                     try:
                         await asyncio.wait_for(task_poll_output_msg, self.iopub_timeout)
                     except (asyncio.TimeoutError, Empty):
@@ -633,10 +754,8 @@ class NotebookClient(LoggingConfigurable):
                 await self._async_handle_timeout(timeout, cell)
 
     async def _async_poll_output_msg(
-            self,
-            parent_msg_id: str,
-            cell: NotebookNode,
-            cell_index: int) -> None:
+        self, parent_msg_id: str, cell: NotebookNode, cell_index: int
+    ) -> None:
 
         assert self.kc is not None
         while True:
@@ -658,7 +777,7 @@ class NotebookClient(LoggingConfigurable):
                 self.task_poll_for_reply.cancel()
                 return
 
-    def _get_timeout(self, cell: t.Optional[NotebookNode]) -> int:
+    def _get_timeout(self, cell: t.Optional[NotebookNode]) -> t.Optional[int]:
         if self.timeout_func is not None and cell is not None:
             timeout = self.timeout_func(cell)
         else:
@@ -670,9 +789,8 @@ class NotebookClient(LoggingConfigurable):
         return timeout
 
     async def _async_handle_timeout(
-            self,
-            timeout: int,
-            cell: t.Optional[NotebookNode] = None) -> None:
+        self, timeout: int, cell: t.Optional[NotebookNode] = None
+    ) -> None:
 
         self.log.error("Timeout waiting for execute reply (%is)." % timeout)
         if self.interrupt_on_timeout:
@@ -691,9 +809,8 @@ class NotebookClient(LoggingConfigurable):
             raise DeadKernelError("Kernel died")
 
     async def async_wait_for_reply(
-            self,
-            msg_id: str,
-            cell: t.Optional[NotebookNode] = None) -> t.Optional[t.Dict]:
+        self, msg_id: str, cell: t.Optional[NotebookNode] = None
+    ) -> t.Optional[t.Dict]:
 
         assert self.kc is not None
         # wait for finish, with timeout
@@ -701,10 +818,8 @@ class NotebookClient(LoggingConfigurable):
         cummulative_time = 0
         while True:
             try:
-                msg = await ensure_async(
-                    self.kc.shell_channel.get_msg(
-                        timeout=self.shell_timeout_interval
-                    )
+                msg: t.Dict = await ensure_async(
+                    self.kc.shell_channel.get_msg(timeout=self.shell_timeout_interval)
                 )
             except Empty:
                 await self._async_check_alive()
@@ -718,7 +833,7 @@ class NotebookClient(LoggingConfigurable):
         return None
 
     wait_for_reply = run_sync(async_wait_for_reply)
-    # Backwards compatability naming for papermill
+    # Backwards compatibility naming for papermill
     _wait_for_reply = wait_for_reply
 
     def _passed_deadline(self, deadline: int) -> bool:
@@ -726,25 +841,33 @@ class NotebookClient(LoggingConfigurable):
             return True
         return False
 
-    def _check_raise_for_error(
-            self,
-            cell: NotebookNode,
-            exec_reply: t.Optional[t.Dict]) -> None:
+    async def _check_raise_for_error(
+        self, cell: NotebookNode, cell_index: int, exec_reply: t.Optional[t.Dict]
+    ) -> None:
 
-        cell_allows_errors = self.allow_errors or "raises-exception" in cell.metadata.get(
-            "tags", []
+        if exec_reply is None:
+            return None
+
+        exec_reply_content = exec_reply['content']
+        if exec_reply_content['status'] != 'error':
+            return None
+
+        cell_allows_errors = (not self.force_raise_errors) and (
+            self.allow_errors
+            or exec_reply_content.get('ename') in self.allow_error_names
+            or "raises-exception" in cell.metadata.get("tags", [])
         )
-
-        if self.force_raise_errors or not cell_allows_errors:
-            if (exec_reply is not None) and exec_reply['content']['status'] == 'error':
-                raise CellExecutionError.from_cell_and_msg(cell, exec_reply['content'])
+        await run_hook(self.on_cell_error, cell=cell, cell_index=cell_index)
+        if not cell_allows_errors:
+            raise CellExecutionError.from_cell_and_msg(cell, exec_reply_content)
 
     async def async_execute_cell(
-            self,
-            cell: NotebookNode,
-            cell_index: int,
-            execution_count: t.Optional[int] = None,
-            store_history: bool = True) -> NotebookNode:
+        self,
+        cell: NotebookNode,
+        cell_index: int,
+        execution_count: t.Optional[int] = None,
+        store_history: bool = True,
+    ) -> NotebookNode:
         """
         Executes a single code cell.
 
@@ -779,21 +902,33 @@ class NotebookClient(LoggingConfigurable):
             The cell which was just processed.
         """
         assert self.kc is not None
+
+        await run_hook(self.on_cell_start, cell=cell, cell_index=cell_index)
+
         if cell.cell_type != 'code' or not cell.source.strip():
             self.log.debug("Skipping non-executing cell %s", cell_index)
             return cell
 
-        if self.record_timing and 'execution' not in cell['metadata']:
+        if self.skip_cells_with_tag in cell.metadata.get("tags", []):
+            self.log.debug("Skipping tagged cell %s", cell_index)
+            return cell
+
+        if self.record_timing:  # clear execution metadata prior to execution
             cell['metadata']['execution'] = {}
 
         self.log.debug("Executing cell:\n%s", cell.source)
+
+        cell_allows_errors = (not self.force_raise_errors) and (
+            self.allow_errors or "raises-exception" in cell.metadata.get("tags", [])
+        )
+
+        await run_hook(self.on_cell_execute, cell=cell, cell_index=cell_index)
         parent_msg_id = await ensure_async(
             self.kc.execute(
-                cell.source,
-                store_history=store_history,
-                stop_on_error=not self.allow_errors
+                cell.source, store_history=store_history, stop_on_error=not cell_allows_errors
             )
         )
+        await run_hook(self.on_cell_complete, cell=cell, cell_index=cell_index)
         # We launched a code cell to execute
         self.code_cells_executed += 1
         exec_timeout = self._get_timeout(cell)
@@ -801,9 +936,7 @@ class NotebookClient(LoggingConfigurable):
         cell.outputs = []
         self.clear_before_next_output = False
 
-        task_poll_kernel_alive = asyncio.ensure_future(
-            self._async_poll_kernel_alive()
-        )
+        task_poll_kernel_alive = asyncio.ensure_future(self._async_poll_kernel_alive())
         task_poll_output_msg = asyncio.ensure_future(
             self._async_poll_output_msg(parent_msg_id, cell, cell_index)
         )
@@ -829,17 +962,15 @@ class NotebookClient(LoggingConfigurable):
 
         if execution_count:
             cell['execution_count'] = execution_count
-        self._check_raise_for_error(cell, exec_reply)
+        await self._check_raise_for_error(cell, cell_index, exec_reply)
         self.nb['cells'][cell_index] = cell
         return cell
 
     execute_cell = run_sync(async_execute_cell)
 
     def process_message(
-            self,
-            msg: t.Dict,
-            cell: NotebookNode,
-            cell_index: int) -> t.Optional[t.List]:
+        self, msg: t.Dict, cell: NotebookNode, cell_index: int
+    ) -> t.Optional[t.List]:
         """
         Processes a kernel message, updates cell state, and returns the
         resulting output object that was appended to cell.outputs.
@@ -882,11 +1013,11 @@ class NotebookClient(LoggingConfigurable):
         if self.record_timing:
             if msg_type == 'status':
                 if content['execution_state'] == 'idle':
-                    cell['metadata']['execution']['iopub.status.idle'] = timestamp()
+                    cell['metadata']['execution']['iopub.status.idle'] = timestamp(msg)
                 elif content['execution_state'] == 'busy':
-                    cell['metadata']['execution']['iopub.status.busy'] = timestamp()
+                    cell['metadata']['execution']['iopub.status.busy'] = timestamp(msg)
             elif msg_type == 'execute_input':
-                cell['metadata']['execution']['iopub.execute_input'] = timestamp()
+                cell['metadata']['execution']['iopub.execute_input'] = timestamp(msg)
 
         if msg_type == 'status':
             if content['execution_state'] == 'idle':
@@ -902,17 +1033,14 @@ class NotebookClient(LoggingConfigurable):
         return None
 
     def output(
-            self,
-            outs: t.List,
-            msg: t.Dict,
-            display_id: str,
-            cell_index: int) -> t.Optional[t.List]:
+        self, outs: t.List, msg: t.Dict, display_id: str, cell_index: int
+    ) -> t.Optional[NotebookNode]:
 
         msg_type = msg['msg_type']
 
         parent_msg_id = msg['parent_header'].get('msg_id')
         if self.output_hook_stack[parent_msg_id]:
-            # if we have a hook registered, it will overrride our
+            # if we have a hook registered, it will override our
             # default output behaviour (e.g. OutputWidget)
             hook = self.output_hook_stack[parent_msg_id][-1]
             hook.output(outs, msg, display_id, cell_index)
@@ -921,7 +1049,7 @@ class NotebookClient(LoggingConfigurable):
         try:
             out = output_from_msg(msg)
         except ValueError:
-            self.log.error("unhandled iopub msg: " + msg_type)
+            self.log.error(f"unhandled iopub msg: {msg_type}")
             return None
 
         if self.clear_before_next_output:
@@ -941,17 +1069,13 @@ class NotebookClient(LoggingConfigurable):
 
         return out
 
-    def clear_output(
-            self,
-            outs: t.List,
-            msg: t.Dict,
-            cell_index: int) -> None:
+    def clear_output(self, outs: t.List, msg: t.Dict, cell_index: int) -> None:
 
         content = msg['content']
 
         parent_msg_id = msg['parent_header'].get('msg_id')
         if self.output_hook_stack[parent_msg_id]:
-            # if we have a hook registered, it will overrride our
+            # if we have a hook registered, it will override our
             # default clear_output behaviour (e.g. OutputWidget)
             hook = self.output_hook_stack[parent_msg_id][-1]
             hook.clear_output(outs, msg, cell_index)
@@ -965,26 +1089,27 @@ class NotebookClient(LoggingConfigurable):
             outs[:] = []
             self.clear_display_id_mapping(cell_index)
 
-    def clear_display_id_mapping(
-            self,
-            cell_index: int) -> None:
+    def clear_display_id_mapping(self, cell_index: int) -> None:
 
-        for display_id, cell_map in self._display_id_map.items():
+        for _, cell_map in self._display_id_map.items():
             if cell_index in cell_map:
                 cell_map[cell_index] = []
 
-    def handle_comm_msg(
-            self,
-            outs: t.List,
-            msg: t.Dict,
-            cell_index: int) -> None:
+    def handle_comm_msg(self, outs: t.List, msg: t.Dict, cell_index: int) -> None:
 
         content = msg['content']
         data = content['data']
         if self.store_widget_state and 'state' in data:  # ignore custom msg'es
             self.widget_state.setdefault(content['comm_id'], {}).update(data['state'])
             if 'buffer_paths' in data and data['buffer_paths']:
-                self.widget_buffers[content['comm_id']] = self._get_buffer_data(msg)
+                comm_id = content['comm_id']
+                if comm_id not in self.widget_buffers:
+                    self.widget_buffers[comm_id] = {}
+                # for each comm, the path uniquely identifies a buffer
+                new_buffers: t.Dict[t.Tuple[str, ...], t.Dict[str, str]] = {
+                    tuple(k["path"]): k for k in self._get_buffer_data(msg)
+                }
+                self.widget_buffers[comm_id].update(new_buffers)
         # There are cases where we need to mimic a frontend, to get similar behaviour as
         # when using the Output widget from Jupyter lab/notebook
         if msg['msg_type'] == 'comm_open':
@@ -1026,10 +1151,7 @@ class NotebookClient(LoggingConfigurable):
             )
         return encoded_buffers
 
-    def register_output_hook(
-            self,
-            msg_id: str,
-            hook: OutputWidget) -> None:
+    def register_output_hook(self, msg_id: str, hook: OutputWidget) -> None:
         """Registers an override object that handles output/clear_output instead.
 
         Multiple hooks can be registered, where the last one will be used (stack based)
@@ -1038,17 +1160,14 @@ class NotebookClient(LoggingConfigurable):
         # https://jupyterlab.github.io/jupyterlab/services/interfaces/kernel.ikernelconnection.html#registermessagehook
         self.output_hook_stack[msg_id].append(hook)
 
-    def remove_output_hook(
-            self,
-            msg_id: str,
-            hook: OutputWidget) -> None:
+    def remove_output_hook(self, msg_id: str, hook: OutputWidget) -> None:
         """Unregisters an override object that handles output/clear_output instead"""
         # mimics
         # https://jupyterlab.github.io/jupyterlab/services/interfaces/kernel.ikernelconnection.html#removemessagehook
         removed_hook = self.output_hook_stack[msg_id].pop()
         assert removed_hook == hook
 
-    def on_comm_open_jupyter_widget(self, msg: t.Dict):
+    def on_comm_open_jupyter_widget(self, msg: t.Dict) -> t.Optional[t.Any]:
         content = msg['content']
         data = content['data']
         state = data['state']
@@ -1058,13 +1177,15 @@ class NotebookClient(LoggingConfigurable):
             widget_class = module.get(state['_model_name'])
             if widget_class:
                 return widget_class(comm_id, state, self.kc, self)
+        return None
 
 
 def execute(
-        nb: NotebookNode,
-        cwd: t.Optional[str] = None,
-        km: t.Optional[KernelManager] = None,
-        **kwargs) -> NotebookClient:
+    nb: NotebookNode,
+    cwd: t.Optional[str] = None,
+    km: t.Optional[KernelManager] = None,
+    **kwargs: t.Any,
+) -> NotebookNode:
     """Execute a notebook's code, updating outputs within the notebook object.
 
     This is a convenient wrapper around NotebookClient. It returns the
