@@ -1,39 +1,50 @@
-from base64 import b64encode, b64decode
+import asyncio
+import concurrent.futures
 import copy
-import io
+import datetime
+import functools
 import os
 import re
 import threading
-import asyncio
-import datetime
 import warnings
-
-import nbformat
-import sys
-import pytest
-import functools
-import xmltodict
-
-from .base import NBClientTestsBase
-from .. import NotebookClient, execute
-from ..exceptions import CellExecutionError
-
-from traitlets import TraitError
-from nbformat import NotebookNode
-from jupyter_client import KernelManager
-from jupyter_client.kernelspec import KernelSpecManager
-from nbconvert.filters import strip_ansi
-from testpath import modified_env
-from ipython_genutils.py3compat import string_types
-import concurrent.futures
-
+from base64 import b64decode, b64encode
 from queue import Empty
 from unittest.mock import MagicMock, Mock
 
+import nbformat
+import pytest
+import xmltodict
+from jupyter_client import KernelManager
+from jupyter_client.kernelspec import KernelSpecManager
+from nbconvert.filters import strip_ansi
+from nbformat import NotebookNode
+from testpath import modified_env
+from traitlets import TraitError
+
+from .. import NotebookClient, execute
+from ..exceptions import CellExecutionError
+from .base import NBClientTestsBase
 
 addr_pat = re.compile(r'0x[0-9a-f]{7,9}')
-ipython_input_pat = re.compile(r'<ipython-input-\d+-[0-9a-f]+>')
 current_dir = os.path.dirname(__file__)
+ipython_input_pat = re.compile(
+    r'(<ipython-input-\d+-[0-9a-f]+>|<IPY-INPUT>) in (<module>|<cell line: \d>\(\))'
+)
+# Tracebacks look different in IPython 8,
+# see: https://github.com/ipython/ipython/blob/master/docs/source/whatsnew/version8.rst#traceback-improvements  # noqa
+ipython8_input_pat = re.compile(
+    r'(Input In \[\d+\]|<IPY-INPUT>), in (<module>|<cell line: \d>\(\))'
+)
+
+hook_methods = [
+    "on_cell_start",
+    "on_cell_execute",
+    "on_cell_complete",
+    "on_cell_error",
+    "on_notebook_start",
+    "on_notebook_complete",
+    "on_notebook_error",
+]
 
 
 class AsyncMock(Mock):
@@ -43,6 +54,7 @@ class AsyncMock(Mock):
 def make_async(mock_value):
     async def _():
         return mock_value
+
     return _()
 
 
@@ -60,7 +72,7 @@ def run_notebook(filename, opts, resources=None):
     running it and the version after running it.
 
     """
-    with io.open(filename) as f:
+    with open(filename) as f:
         input_nb = nbformat.read(f, 4)
 
     cleaned_input_nb = copy.deepcopy(input_nb)
@@ -94,7 +106,7 @@ async def async_run_notebook(filename, opts, resources=None):
     running it and the version after running it.
 
     """
-    with io.open(filename) as f:
+    with open(filename) as f:
         input_nb = nbformat.read(f, 4)
 
     cleaned_input_nb = copy.deepcopy(input_nb)
@@ -114,7 +126,7 @@ async def async_run_notebook(filename, opts, resources=None):
     return input_nb, output_nb
 
 
-def prepare_cell_mocks(*messages, reply_msg=None):
+def prepare_cell_mocks(*messages_input, reply_msg=None):
     """
     This function prepares a executor object which has a fake kernel client
     to mock the messages sent over zeromq. The mock kernel client will return
@@ -122,7 +134,7 @@ def prepare_cell_mocks(*messages, reply_msg=None):
     callbacks. It also appends a kernel idle message to the end of messages.
     """
     parent_id = 'fake_id'
-    messages = list(messages)
+    messages = list(messages_input)
     # Always terminate messages with an idle to exit the loop
     messages.append({'msg_type': 'status', 'content': {'execution_state': 'idle'}})
 
@@ -130,13 +142,15 @@ def prepare_cell_mocks(*messages, reply_msg=None):
         # Return the message generator for
         # self.kc.shell_channel.get_msg => {'parent_header': {'msg_id': parent_id}}
         return AsyncMock(
-            return_value=make_async(NBClientTestsBase.merge_dicts(
-                {
-                    'parent_header': {'msg_id': parent_id},
-                    'content': {'status': 'ok', 'execution_count': 1},
-                },
-                reply_msg or {},
-            ))
+            return_value=make_async(
+                NBClientTestsBase.merge_dicts(
+                    {
+                        'parent_header': {'msg_id': parent_id},
+                        'content': {'status': 'ok', 'execution_count': 1},
+                    },
+                    reply_msg or {},
+                )
+            )
         )
 
     def iopub_messages_mock():
@@ -174,7 +188,7 @@ def prepare_cell_mocks(*messages, reply_msg=None):
                 iopub_channel=MagicMock(get_msg=message_mock),
                 shell_channel=MagicMock(get_msg=shell_channel_message_mock()),
                 execute=MagicMock(return_value=parent_id),
-                is_alive=MagicMock(return_value=make_async(True))
+                is_alive=MagicMock(return_value=make_async(True)),
             )
             executor.parent_id = parent_id
             return func(self, executor, cell_mock, message_mock)
@@ -200,13 +214,14 @@ def normalize_output(output):
     if 'image/svg+xml' in output.get('data', {}):
         output['data']['image/svg+xml'] = xmltodict.parse(output['data']['image/svg+xml'])
     for key, value in output.get('data', {}).items():
-        if isinstance(value, string_types):
+        if isinstance(value, str):
             output['data'][key] = normalize_base64(value)
     if 'traceback' in output:
-        tb = [
-            re.sub(ipython_input_pat, '<IPY-INPUT>', strip_ansi(line))
-            for line in output['traceback']
-        ]
+        tb = []
+        for line in output["traceback"]:
+            line = re.sub(ipython_input_pat, '<IPY-INPUT>', strip_ansi(line))
+            line = re.sub(ipython8_input_pat, '<IPY-INPUT>', strip_ansi(line))
+            tb.append(line)
         output['traceback'] = tb
 
     return output
@@ -241,6 +256,16 @@ def notebook_resources():
     return {'metadata': {'path': os.path.join(current_dir, 'files')}}
 
 
+def filter_messages_on_error_output(err_output):
+    allowed_lines = [
+        # ipykernel migh be installed without debugpy extension
+        "[IPKernelApp] WARNING | debugpy_stream undefined, debugging will not be enabled",
+    ]
+    filtered_result = [line for line in err_output.splitlines() if line not in allowed_lines]
+
+    return os.linesep.join(filtered_result)
+
+
 @pytest.mark.parametrize(
     ["input_name", "opts"],
     [
@@ -257,6 +282,7 @@ def notebook_resources():
         ("JupyterWidgets.ipynb", dict(kernel_name="python")),
         ("Skip Exceptions with Cell Tags.ipynb", dict(kernel_name="python")),
         ("Skip Exceptions.ipynb", dict(kernel_name="python", allow_errors=True)),
+        ("Skip Execution with Cell Tag.ipynb", dict(kernel_name="python")),
         ("SVG.ipynb", dict(kernel_name="python")),
         ("Unicode.ipynb", dict(kernel_name="python")),
         ("UnicodePy3.ipynb", dict(kernel_name="python")),
@@ -287,11 +313,13 @@ def test_parallel_notebooks(capfd, tmpdir):
             threading.Thread(target=run_notebook, args=(input_file.format(label=label), opts, res))
             for label in ("A", "B")
         ]
-        [t.start() for t in threads]
-        [t.join(timeout=2) for t in threads]
+        for t in threads:
+            t.start()
+        for t in threads:
+            t.join(timeout=2)
 
     captured = capfd.readouterr()
-    assert captured.err == ""
+    assert filter_messages_on_error_output(captured.err) == ""
 
 
 def test_many_parallel_notebooks(capfd):
@@ -317,7 +345,7 @@ def test_many_parallel_notebooks(capfd):
             executor.map(run_notebook_wrapper, [(input_file, opts, res) for i in range(8)])
 
     captured = capfd.readouterr()
-    assert captured.err == ""
+    assert filter_messages_on_error_output(captured.err) == ""
 
 
 def test_async_parallel_notebooks(capfd, tmpdir):
@@ -333,14 +361,13 @@ def test_async_parallel_notebooks(capfd, tmpdir):
 
     with modified_env({"NBEXECUTE_TEST_PARALLEL_TMPDIR": str(tmpdir)}):
         tasks = [
-            async_run_notebook(input_file.format(label=label), opts, res)
-            for label in ("A", "B")
+            async_run_notebook(input_file.format(label=label), opts, res) for label in ("A", "B")
         ]
         loop = asyncio.get_event_loop()
         loop.run_until_complete(asyncio.gather(*tasks))
 
     captured = capfd.readouterr()
-    assert captured.err == ""
+    assert filter_messages_on_error_output(captured.err) == ""
 
 
 def test_many_async_parallel_notebooks(capfd):
@@ -358,15 +385,12 @@ def test_many_async_parallel_notebooks(capfd):
     # run once, to trigger creating the original context
     run_notebook(input_file, opts, res)
 
-    tasks = [
-        async_run_notebook(input_file, opts, res)
-        for i in range(4)
-    ]
+    tasks = [async_run_notebook(input_file, opts, res) for i in range(4)]
     loop = asyncio.get_event_loop()
     loop.run_until_complete(asyncio.gather(*tasks))
 
     captured = capfd.readouterr()
-    assert captured.err == ""
+    assert filter_messages_on_error_output(captured.err) == ""
 
 
 def test_execution_timing():
@@ -403,7 +427,7 @@ def test_synchronous_setup_kernel():
     nb = nbformat.v4.new_notebook()
     executor = NotebookClient(nb)
     with executor.setup_kernel():
-        # Prove it initalized client
+        # Prove it initialized client
         assert executor.kc is not None
     # Prove it removed the client (and hopefully cleaned up)
     assert executor.kc is None
@@ -415,12 +439,12 @@ def test_startnewkernel_with_kernelmanager():
     executor = NotebookClient(nb, km=km)
     executor.start_new_kernel()
     kc = executor.start_new_kernel_client()
-    # prove it initalized client
+    # prove it initialized client
     assert kc is not None
     # since we are not using the setup_kernel context manager,
     # cleanup has to be done manually
     kc.shutdown()
-    km.cleanup()
+    km.cleanup_resources()
     kc.stop_channels()
 
 
@@ -442,7 +466,7 @@ def test_start_new_kernel_history_file_setting():
     # since we are not using the setup_kernel context manager,
     # cleanup has to be done manually
     kc.shutdown()
-    km.cleanup()
+    km.cleanup_resources()
     kc.stop_channels()
 
 
@@ -540,7 +564,7 @@ while True: continue
     def test_kernel_death_after_timeout(self):
         """Check that an error is raised when the kernel is_alive is false after a cell timed out"""
         filename = os.path.join(current_dir, 'files', 'Interrupt.ipynb')
-        with io.open(filename, 'r') as f:
+        with open(filename) as f:
             input_nb = nbformat.read(f, 4)
         res = self.build_resources()
         res['metadata']['path'] = os.path.dirname(filename)
@@ -553,6 +577,7 @@ while True: continue
 
         async def is_alive():
             return False
+
         km.is_alive = is_alive
         # Will be a RuntimeError or subclass DeadKernelError depending
         # on if jupyter_client or nbconvert catches the dead client first
@@ -564,7 +589,7 @@ while True: continue
         execution.
         """
         filename = os.path.join(current_dir, 'files', 'Autokill.ipynb')
-        with io.open(filename, 'r') as f:
+        with open(filename) as f:
             input_nb = nbformat.read(f, 4)
 
         executor = NotebookClient(input_nb)
@@ -582,10 +607,7 @@ while True: continue
         with pytest.raises(CellExecutionError) as exc:
             run_notebook(filename, dict(allow_errors=False), res)
             self.assertIsInstance(str(exc.value), str)
-            if sys.version_info >= (3, 0):
-                assert u"# üñîçø∂é" in str(exc.value)
-            else:
-                assert u"# üñîçø∂é".encode('utf8', 'replace') in str(exc.value)
+            assert "# üñîçø∂é" in str(exc.value)
 
     def test_force_raise_errors(self):
         """
@@ -598,15 +620,12 @@ while True: continue
         with pytest.raises(CellExecutionError) as exc:
             run_notebook(filename, dict(force_raise_errors=True), res)
             self.assertIsInstance(str(exc.value), str)
-            if sys.version_info >= (3, 0):
-                assert u"# üñîçø∂é" in str(exc.value)
-            else:
-                assert u"# üñîçø∂é".encode('utf8', 'replace') in str(exc.value)
+            assert "# üñîçø∂é" in str(exc.value)
 
     def test_reset_kernel_client(self):
         filename = os.path.join(current_dir, 'files', 'HelloWorld.ipynb')
 
-        with io.open(filename) as f:
+        with open(filename) as f:
             input_nb = nbformat.read(f, 4)
 
         executor = NotebookClient(
@@ -631,7 +650,7 @@ while True: continue
     def test_cleanup_kernel_client(self):
         filename = os.path.join(current_dir, 'files', 'HelloWorld.ipynb')
 
-        with io.open(filename) as f:
+        with open(filename) as f:
             input_nb = nbformat.read(f, 4)
 
         executor = NotebookClient(
@@ -653,7 +672,7 @@ while True: continue
 
         filename = os.path.join(current_dir, 'files', 'HelloWorld.ipynb')
 
-        with io.open(filename) as f:
+        with open(filename) as f:
             input_nb = nbformat.read(f, 4)
 
         cleaned_input_nb = copy.deepcopy(input_nb)
@@ -675,14 +694,14 @@ while True: continue
         expected = FakeCustomKernelManager.expected_methods.items()
 
         for method, call_count in expected:
-            self.assertNotEqual(call_count, 0, '{} was called'.format(method))
+            self.assertNotEqual(call_count, 0, f'{method} was called')
 
     def test_process_message_wrapper(self):
-        outputs = []
+        outputs: list = []
 
         class WrappedPreProc(NotebookClient):
             def process_message(self, msg, cell, cell_index):
-                result = super(WrappedPreProc, self).process_message(msg, cell, cell_index)
+                result = super().process_message(msg, cell, cell_index)
                 if result:
                     outputs.append(result)
                 return result
@@ -690,7 +709,7 @@ while True: continue
         current_dir = os.path.dirname(__file__)
         filename = os.path.join(current_dir, 'files', 'HelloWorld.ipynb')
 
-        with io.open(filename) as f:
+        with open(filename) as f:
             input_nb = nbformat.read(f, 4)
 
         original = copy.deepcopy(input_nb)
@@ -703,7 +722,7 @@ while True: continue
         # Test the execute() convenience API
         filename = os.path.join(current_dir, 'files', 'HelloWorld.ipynb')
 
-        with io.open(filename) as f:
+        with open(filename) as f:
             input_nb = nbformat.read(f, 4)
 
         original = copy.deepcopy(input_nb)
@@ -737,6 +756,82 @@ while True: continue
         assert 'version_major' in wdata
         assert 'version_minor' in wdata
 
+    def test_execution_hook(self):
+        filename = os.path.join(current_dir, 'files', 'HelloWorld.ipynb')
+        with open(filename) as f:
+            input_nb = nbformat.read(f, 4)
+        hooks = [MagicMock() for i in range(7)]
+        executor = NotebookClient(input_nb)
+        for executor_hook, hook in zip(hook_methods, hooks):
+            setattr(executor, executor_hook, hook)
+        executor.execute()
+        for hook in hooks[:3]:
+            hook.assert_called_once()
+        hooks[3].assert_not_called()
+        for hook in hooks[4:6]:
+            hook.assert_called_once()
+        hooks[6].assert_not_called()
+
+    def test_error_execution_hook_error(self):
+        filename = os.path.join(current_dir, 'files', 'Error.ipynb')
+        with open(filename) as f:
+            input_nb = nbformat.read(f, 4)
+        hooks = [MagicMock() for i in range(7)]
+        executor = NotebookClient(input_nb)
+        for executor_hook, hook in zip(hook_methods, hooks):
+            setattr(executor, executor_hook, hook)
+        with pytest.raises(CellExecutionError):
+            executor.execute()
+        for hook in hooks[:5]:
+            hook.assert_called_once()
+        hooks[6].assert_not_called()
+
+    def test_error_notebook_hook(self):
+        filename = os.path.join(current_dir, 'files', 'Autokill.ipynb')
+        with open(filename) as f:
+            input_nb = nbformat.read(f, 4)
+        hooks = [MagicMock() for i in range(7)]
+        executor = NotebookClient(input_nb)
+        for executor_hook, hook in zip(hook_methods, hooks):
+            setattr(executor, executor_hook, hook)
+        with pytest.raises(RuntimeError):
+            executor.execute()
+        for hook in hooks[:3]:
+            hook.assert_called_once()
+        hooks[3].assert_not_called()
+        for hook in hooks[4:]:
+            hook.assert_called_once()
+
+    def test_async_execution_hook(self):
+        filename = os.path.join(current_dir, 'files', 'HelloWorld.ipynb')
+        with open(filename) as f:
+            input_nb = nbformat.read(f, 4)
+        hooks = [AsyncMock() for i in range(7)]
+        executor = NotebookClient(input_nb)
+        for executor_hook, hook in zip(hook_methods, hooks):
+            setattr(executor, executor_hook, hook)
+        executor.execute()
+        for hook in hooks[:3]:
+            hook.assert_called_once()
+        hooks[3].assert_not_called()
+        for hook in hooks[4:6]:
+            hook.assert_called_once()
+        hooks[6].assert_not_called()
+
+    def test_error_async_execution_hook(self):
+        filename = os.path.join(current_dir, 'files', 'Error.ipynb')
+        with open(filename) as f:
+            input_nb = nbformat.read(f, 4)
+        hooks = [AsyncMock() for i in range(7)]
+        executor = NotebookClient(input_nb)
+        for executor_hook, hook in zip(hook_methods, hooks):
+            setattr(executor, executor_hook, hook)
+        with pytest.raises(CellExecutionError):
+            executor.execute().execute()
+        for hook in hooks[:5]:
+            hook.assert_called_once()
+        hooks[6].assert_not_called()
+
 
 class TestRunCell(NBClientTestsBase):
     """Contains test functions for NotebookClient.execute_cell"""
@@ -836,10 +931,10 @@ class TestRunCell(NBClientTestsBase):
     def test_eventual_deadline_iopub(self, executor, cell_mock, message_mock):
         # Process a few messages before raising a timeout from iopub
         def message_seq(messages):
-            for message in messages:
-                yield message
+            yield from messages
             while True:
                 yield Empty()
+
         message_mock.side_effect = message_seq(list(message_mock.side_effect)[:-1])
         executor.kc.shell_channel.get_msg = Mock(
             return_value=make_async({'parent_header': {'msg_id': executor.parent_id}})
@@ -1051,17 +1146,51 @@ class TestRunCell(NBClientTestsBase):
             'buffers': [b'123'],
             'content': {
                 'comm_id': 'foobar',
-                'data': {'state': {'foo': 'bar'}, 'buffer_paths': ['path']},
+                'data': {'state': {'foo': 'bar'}, 'buffer_paths': [['path']]},
             },
         }
     )
-    def test_widget_comm_buffer_message(self, executor, cell_mock, message_mock):
+    def test_widget_comm_buffer_message_single(self, executor, cell_mock, message_mock):
         executor.execute_cell(cell_mock, 0)
         # A comm message with buffer info followed by an idle
         assert message_mock.call_count == 2
         assert executor.widget_state == {'foobar': {'foo': 'bar'}}
         assert executor.widget_buffers == {
-            'foobar': [{'data': 'MTIz', 'encoding': 'base64', 'path': 'path'}]
+            'foobar': {('path',): {'data': 'MTIz', 'encoding': 'base64', 'path': ['path']}}
+        }
+        # Ensure no outputs were generated
+        assert cell_mock.outputs == []
+
+    @prepare_cell_mocks(
+        {
+            'msg_type': 'comm',
+            'header': {'msg_type': 'comm'},
+            'buffers': [b'123'],
+            'content': {
+                'comm_id': 'foobar',
+                'data': {'state': {'foo': 'bar'}, 'buffer_paths': [['path']]},
+            },
+        },
+        {
+            'msg_type': 'comm',
+            'header': {'msg_type': 'comm'},
+            'buffers': [b'123'],
+            'content': {
+                'comm_id': 'foobar',
+                'data': {'state': {'foo2': 'bar2'}, 'buffer_paths': [['path2']]},
+            },
+        },
+    )
+    def test_widget_comm_buffer_messages(self, executor, cell_mock, message_mock):
+        executor.execute_cell(cell_mock, 0)
+        # A comm message with buffer info followed by an idle
+        assert message_mock.call_count == 3
+        assert executor.widget_state == {'foobar': {'foo': 'bar', 'foo2': 'bar2'}}
+        assert executor.widget_buffers == {
+            'foobar': {
+                ('path',): {'data': 'MTIz', 'encoding': 'base64', 'path': ['path']},
+                ('path2',): {'data': 'MTIz', 'encoding': 'base64', 'path': ['path2']},
+            }
         }
         # Ensure no outputs were generated
         assert cell_mock.outputs == []
@@ -1409,6 +1538,24 @@ class TestRunCell(NBClientTestsBase):
         # Should also consume the message stream
         assert cell_mock.outputs == []
 
+    @prepare_cell_mocks(
+        reply_msg={
+            'msg_type': 'execute_reply',
+            'header': {'msg_type': 'execute_reply'},
+            # ERROR
+            'content': {'status': 'error', 'ename': 'NotImplementedError'},
+        }
+    )
+    def test_allow_error_names(self, executor, cell_mock, message_mock):
+        executor.allow_error_names = ['NotImplementedError']
+        # Should NOT raise
+        executor.execute_cell(cell_mock, 0)
+
+        # An error followed by an idle
+        assert message_mock.call_count == 1
+        # Should also consume the message stream
+        assert cell_mock.outputs == []
+
     @prepare_cell_mocks(
         reply_msg={
             'msg_type': 'execute_reply',
@@ -1468,3 +1615,92 @@ class TestRunCell(NBClientTestsBase):
         assert message_mock.call_count == 0
         # Should also consume the message stream
         assert cell_mock.outputs == []
+
+    @prepare_cell_mocks()
+    def test_cell_hooks(self, executor, cell_mock, message_mock):
+        hooks = [MagicMock() for i in range(7)]
+        for executor_hook, hook in zip(hook_methods, hooks):
+            setattr(executor, executor_hook, hook)
+        executor.execute_cell(cell_mock, 0)
+        for hook in hooks[:3]:
+            hook.assert_called_once_with(cell=cell_mock, cell_index=0)
+        for hook in hooks[4:]:
+            hook.assert_not_called()
+
+    @prepare_cell_mocks(
+        {
+            'msg_type': 'error',
+            'header': {'msg_type': 'error'},
+            'content': {'ename': 'foo', 'evalue': 'bar', 'traceback': ['Boom']},
+        },
+        reply_msg={
+            'msg_type': 'execute_reply',
+            'header': {'msg_type': 'execute_reply'},
+            # ERROR
+            'content': {'status': 'error'},
+        },
+    )
+    def test_error_cell_hooks(self, executor, cell_mock, message_mock):
+        hooks = [MagicMock() for i in range(7)]
+        for executor_hook, hook in zip(hook_methods, hooks):
+            setattr(executor, executor_hook, hook)
+        with self.assertRaises(CellExecutionError):
+            executor.execute_cell(cell_mock, 0)
+        for hook in hooks[:4]:
+            hook.assert_called_once_with(cell=cell_mock, cell_index=0)
+        for hook in hooks[5:]:
+            hook.assert_not_called()
+
+    @prepare_cell_mocks(
+        reply_msg={
+            'msg_type': 'execute_reply',
+            'header': {'msg_type': 'execute_reply'},
+            # ERROR
+            'content': {'status': 'error'},
+        }
+    )
+    def test_non_code_cell_hooks(self, executor, cell_mock, message_mock):
+        cell_mock = NotebookNode(source='"foo" = "bar"', metadata={}, cell_type='raw', outputs=[])
+        hooks = [MagicMock() for i in range(7)]
+        for executor_hook, hook in zip(hook_methods, hooks):
+            setattr(executor, executor_hook, hook)
+        executor.execute_cell(cell_mock, 0)
+        for hook in hooks[:1]:
+            hook.assert_called_once_with(cell=cell_mock, cell_index=0)
+        for hook in hooks[1:]:
+            hook.assert_not_called()
+
+    @prepare_cell_mocks()
+    def test_async_cell_hooks(self, executor, cell_mock, message_mock):
+        hooks = [AsyncMock() for i in range(7)]
+        for executor_hook, hook in zip(hook_methods, hooks):
+            setattr(executor, executor_hook, hook)
+        executor.execute_cell(cell_mock, 0)
+        for hook in hooks[:3]:
+            hook.assert_called_once_with(cell=cell_mock, cell_index=0)
+        for hook in hooks[4:]:
+            hook.assert_not_called()
+
+    @prepare_cell_mocks(
+        {
+            'msg_type': 'error',
+            'header': {'msg_type': 'error'},
+            'content': {'ename': 'foo', 'evalue': 'bar', 'traceback': ['Boom']},
+        },
+        reply_msg={
+            'msg_type': 'execute_reply',
+            'header': {'msg_type': 'execute_reply'},
+            # ERROR
+            'content': {'status': 'error'},
+        },
+    )
+    def test_error_async_cell_hooks(self, executor, cell_mock, message_mock):
+        hooks = [AsyncMock() for i in range(7)]
+        for executor_hook, hook in zip(hook_methods, hooks):
+            setattr(executor, executor_hook, hook)
+        with self.assertRaises(CellExecutionError):
+            executor.execute_cell(cell_mock, 0)
+        for hook in hooks[:4]:
+            hook.assert_called_once_with(cell=cell_mock, cell_index=0)
+        for hook in hooks[4:]:
+            hook.assert_not_called()
